import java.math.BigInteger;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.MessageDigest;
import java.security.Security;
import java.security.cert.X509Certificate;
import java.util.Date;

import org.bouncycastle.asn1.x500.X500Name;
import org.bouncycastle.asn1.x509.GeneralName;
import org.bouncycastle.cert.X509CertificateHolder;
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter;
import org.bouncycastle.cert.jcajce.JcaX509v1CertificateBuilder;
import org.bouncycastle.cert.jcajce.JcaX509v3CertificateBuilder;
import org.bouncycastle.jcajce.ExternalPublicKey;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder;

public class ExternalKeyCertificateExample
{
    private static long ONE_YEAR = 365 * 24 * 60 * 60 * 1000L;

    public static void main(String[] args)
        throws Exception
    {
        Security.addProvider(new BouncyCastleProvider());

        // KeyPairGenerator kpGen = KeyPairGenerator.getInstance("ML-DSA-87");

        KeyPair kp = kpGen.generateKeyPair();

        // ExternalPublicKey externalKey = new ExternalPublicKey(kp.getPublic(),
        //   new GeneralName(GeneralName.uniformResourceIdentifier, "https://localhost"),
        //   MessageDigest.getInstance("SHA256"));

        X500Name name = new X500Name("CN=ML-DSA-87 External Key Example");
        long time = System.currentTimeMillis();
        JcaX509v3CertificateBuilder certBldr = new JcaX509v3CertificateBuilder(
            name, BigInteger.valueOf(System.currentTimeMillis()), new Date(time - 5000), new Date(time + ONE_YEAR),
            name, externalKey);

        // X509CertificateHolder certHolder = certBldr.build(new JcaContentSignerBuilder("ML-DSA-87").build(kp.getPrivate()));

        X509Certificate cert = new JcaX509CertificateConverter().setProvider("BC").getCertificate(certHolder);

        try
        {
            cert.verify(kp.getPublic());

            System.out.println("external public key certificate verified");
            System.exit(0);
        }
        catch (Exception e)
        {
            System.exit(1);
        }
    }
}
