import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.SecureRandom;
import java.security.Security;

import javax.crypto.KeyAgreement;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;

import org.bouncycastle.jcajce.SecretKeyWithEncapsulation;
import org.bouncycastle.jcajce.spec.HybridValueParameterSpec;
import org.bouncycastle.jcajce.spec.KEMExtractSpec;
import org.bouncycastle.jcajce.spec.KEMGenerateSpec;
import org.bouncycastle.jcajce.spec.MLKEMParameterSpec;
import org.bouncycastle.jcajce.spec.UserKeyingMaterialSpec;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.bouncycastle.util.Arrays;
import org.bouncycastle.util.encoders.Hex;

public class SP80056CHybridEncryptionExample
{
    public static void main(String[] args)
        throws Exception
    {
        Security.addProvider(new BouncyCastleProvider());
        byte[] ukm = Hex.decode("030f136fa7fef90d185655ed1c6d46bacdb820");

        KeyPairGenerator kpg = KeyPairGenerator.getInstance("EC", "BC");

        kpg.initialize(256);

        KeyPair kp1 = kpg.generateKeyPair();
        KeyPair kp2 = kpg.generateKeyPair();

        // generate ML-KEM-512 key pair.
        // KeyPairGenerator pqKpg = KeyPairGenerator.getInstance("ML-KEM", "BC");

        // pqKpg.initialize(MLKEMParameterSpec.ml_kem_512, new SecureRandom());

        // KeyPair pqKp = pqKpg.generateKeyPair();

        // Create the KeyGenerator - there are specific specs for creating encapsulations (KEMGenerateSpec) and for
        // extract a secret key from an encapsulation using the private key (KEMExtractSpec)
        // KeyGenerator keyGen = KeyGenerator.getInstance("ML-KEM", "BC");

        // initialise for creating an encapsulation and shared secret.
        // keyGen.init(new KEMGenerateSpec.Builder(pqKp.getPublic(), "Generic", 128)
            // .withNoKdf().build(), new SecureRandom());

        // SecretKeyWithEncapsulation is the class to use as the secret key, it has additional
        // methods on it for recovering the encapsulation as well.
        // SecretKeyWithEncapsulation secEnc1 = (SecretKeyWithEncapsulation)keyGen.generateKey();


        KeyAgreement agreement = KeyAgreement.getInstance("ECCDHwithSHA256CKDF", "BC");

        // agreement.init(kp1.getPrivate(), new HybridValueParameterSpec(secEnc1.getEncoded(), new UserKeyingMaterialSpec(ukm)));

        agreement.doPhase(kp2.getPublic(), true);

        SecretKey k1 = agreement.generateSecret("AES[256]");

        // keyGen.init(new KEMExtractSpec.Builder(pqKp.getPrivate(), secEnc1.getEncapsulation(), "AES", 128).withNoKdf().build());

        // initialise for extracting the shared secret from the encapsulation.
        // SecretKeyWithEncapsulation secEnc2 = (SecretKeyWithEncapsulation)keyGen.generateKey();

        // agreement.init(kp2.getPrivate(), new HybridValueParameterSpec(secEnc2.getEncoded(), new UserKeyingMaterialSpec(ukm)));

        agreement.doPhase(kp1.getPublic(), true);

        SecretKey k2 = agreement.generateSecret("AES[256]");

        // a quick check to make sure we got the same answer on both sides.
        if (Arrays.areEqual(k1.getEncoded(), k2.getEncoded()))
        {
            System.out.println("AES key generated successfully: " + Hex.toHexString(k1.getEncoded()));
            System.exit(0);
        }

        System.exit(1);
    }
}
